%NEWTONANALYSIS 
% Takes clicks statistics and finds optimum values of l1,l2,b1,b2 (the loss
% parameters and background parameters) by using Newton's method.

% An initial guess is provided (e.g. losses ~0.9, background ~0.1), and
% then Newton canes it out by calculating local derivatives and deciding a
% step direction, d, in which to go to find the next point.

% The parameters [l1,l2,b1,b2] will sometimes be referred to as
% [x1,x2,x3,x4]

% Max number of photons to consider i.e. size of state
N = 8;

% Distances in parameter space for calculating derivatives
% h = [h(l1), h(l2), h(b1), h(b2)]
h = [0.01, 0.01]';
gamma = 1;

Imax = 150;
iter = 1;

d = zeros(2,Imax);

dump = zeros(N+1,1); % The minimised states are irrelevant

% Starting vector
x_0 = [0.1,0.1]';

x = zeros(3, Imax);
x(1:2,1) = x_0;

T = zeros(3,7);
D = zeros(7);

G = zeros(2,1);
H = zeros(2,2);

% Find statistics from click counts
P = click/sum(sum(click));
M = zeros(81,N+1);

% Use all P elements to make p an 81x1 vector
prow = 1;
for n = 1:9
    for m = 1:9
        p(prow,1) = P(n,m);
        prow = prow + 1;
    end
end

for iter = 1:Imax
    
    % ---- Calculating the function points for derivatives ----

    % Temporary 5x21 matrix, T, stores these points and their residuals
    %           - T(5,1) = 1 function value at X(1:4,iter)
    %           - T(5,2:9) = 8 function values for single variable X(1:4,iter) +/- h
    %           - T(5,10:21) = 12 function values for cross variables X(1:4,iter) +/- h
    T(1:2,1) = x(1:2,iter); % Load point

    T(1:2,2) = x(1:2,iter) + [h(1),0]'; % Load single variable points
    T(1:2,3) = x(1:2,iter) - [h(1),0]';
    T(1:2,4) = x(1:2,iter) + [0,h(2)]';
    T(1:2,5) = x(1:2,iter) - [0,h(2)]';


    T(1:2,6) = x(1:2,iter) + [h(1), h(2)]'; % Load cross variable points
    T(1:2,7) = x(1:2,iter) + [h(1),-h(2)]';


    for u=1:7
        % Calculate background matrices
        B1 = Bground(T(1,u),N); B2 = Bground(T(2,u),N);

        % Calculate working matrices - initially need P = Q*rho*R
        Q = P_kn1*B1;
        R = B2'*P_kn2';

        % Create 81x(N+1) M matrix - to convert to form p = Mx
        mrow = 1;
        for n = 1:9
            for m = 1:9
            M(mrow,:) = Q(n,:).*R(:,m)';    
            mrow = mrow + 1;
            end
        end

        % Do the least squares minimisation (i.e. compute the value of the
        % function!)
        [dump, T(3,u)] = lsqnonneg(M,p);
        if u==1
            x(3,iter) = T(3,u);
        end
    
    end


     % ---- Calculating the derivative values ----
    % A 14-vector, D, for 
    %           - D(1:4) = 1st derivatives (1,2,3,4)
    %           - D(5:8) = single-variable 2nd derivatives (1,2,3,4)
    %           - D(9:14) = cross-variable 2nd derivatives (12,13,14,23,24,34)

    % 1st derivatives
    D(1) = (T(3,2) - T(3,3))/(2*h(1));
    D(2) = (T(3,4) - T(3,5))/(2*h(2));


    % Single variable 2nd derivatives
    D(5) = (T(3,2) + T(3,3) - T(3,1))/(h(1)^2);
    D(6) = (T(3,4) + T(3,5) - T(3,1))/(h(2)^2);


    % Cross variable 2nd derivatives
    D(7) = (T(3,6) - T(3,7))/(2*h(1)*h(2)) - D(2)/h(1);



     % ---- Constructing the Gradient Vector ----
     G = zeros(2,1);
     G = [D(1),D(2)]';


     % ---- Constructing the Hessian matrix ----
     H = zeros(2,2);
     H(1,1) = D(5); H(2,2) = D(6);
     H(1,2) = D(7);
     H(2,1) = H(1,2);


     % ---- Moving to a new point in parameter space ----
     % Points are recorded in 5 x Imax matrix, X, where Imax = max number of iterations
     if iter < Imax
        d(1:2,iter) = H^(-1)*G;
        x(1:2,iter+1) = x(1:2,iter) - gamma*d(1:2,iter);
     end

     iter
 
 end
 
 
 