% Creator: Andy Worsley
% Dept. of Physics, Univeristy of Oxford
% March 2008

% DESCRIPTION
% -----------
% Takes clicks statistics and finds optimum values of l1,l2 (the loss
% parameters) by using Newton's method to minimise the residual between the
% measured click statistics and those calculated froma non-negative
% least-squares fitted state

% An initial guess is provided (e.g. losses ~0.9, background ~0.1), and
% then the program calculates local derivatives (approximating tangents)
% and decides a step direction, d, in which to go to find the next point.

% INPUTS
% P_kn1 - the convolution matrix for LOOPY 1
% P_kn2 - the convolution matrix for LOOPY 2
% click - the click statistics

% OUTPUTS
% losses_bestrn - vector containing the loss parms at Imax
% x = [x1,x2,x3], where:
%      x1 - loss parameter 1
%      x2 - loss parameter 2
%      x3 - residual at that point in parameter space

% Max number of photons to consider i.e. size of state
N = size(P_kn1,2) - 1; % Read highest considered photon pair number from convolution matrix size





% ---- USER DEFINED PARAMETERS ----
h = [0.001, 0.001]'; % Distances in parameter space over which to numerically calculate derivatives
gamma = 2; % Multiplication factor for direction vector
x_0 = [0.850,0.875]'; % An initial guess is provided for the loss parameters
Imax = 200; % Maximum number of iterations
% ---------------------------------

iter = 1;
d = zeros(2,Imax);
dump = zeros(N+1,1); % The minimised states are irrelevant

x = zeros(3, Imax);
x(1:2,1) = x_0;

T = zeros(3,7);
D = zeros(7);

G = zeros(2,1);
H = zeros(2,2);

% - ERRORS -
% Check P_kn1 and P_kn2 are the same size
if ((size(P_kn1,1)~=size(P_kn2,1)) + (size(P_kn1,2)~=size(P_kn2,2)))
    error('Convolution matrices must be the same size');
end
% ---------

% Find statistics from click counts
P = click/sum(sum(click));
M = zeros(81,N+1);


% ---------------
% ---- START ----
% ---------------

% Use all P elements to make p an 81x1 vector
prow = 1;
for n = 1:9
    for m = 1:9
        p(prow,1) = P(n,m);
        prow = prow + 1;
    end
end

for iter = 1:Imax
    
    % ---- Calculating the function points for derivatives ----

    % Temporary 5x21 matrix, T, stores these points and their residuals
    %           - T(5,1) = 1 function value at X(1:4,iter)
    %           - T(5,2:9) = 8 function values for single variable X(1:4,iter) +/- h
    %           - T(5,10:21) = 12 function values for cross variables X(1:4,iter) +/- h
    T(1:2,1) = x(1:2,iter); % Load point

    T(1:2,2) = x(1:2,iter) + [h(1),0]'; % Load single variable points
    T(1:2,3) = x(1:2,iter) - [h(1),0]';
    T(1:2,4) = x(1:2,iter) + [0,h(2)]';
    T(1:2,5) = x(1:2,iter) - [0,h(2)]';


    T(1:2,6) = x(1:2,iter) + [h(1), h(2)]'; % Load cross variable points
    T(1:2,7) = x(1:2,iter) + [h(1),-h(2)]';


    for u=1:7
        % Calculate loss matrices
        L1 = Loss(T(1,u),N); L2 = Loss(T(2,u),N);

        % Calculate working matrices - initially need P = Q*rho*R
        Q = P_kn1*L1;
        R = L2'*P_kn2';

        % Create 81x(N+1) M matrix - to convert to form p = Mx
        mrow = 1;
        for n = 1:9
            for m = 1:9
            M(mrow,:) = Q(n,:).*R(:,m)';    
            mrow = mrow + 1;
            end
        end

        % Do the least squares minimisation (i.e. compute the value of the
        % function!)
        [dump, T(3,u)] = lsqnonneg(M,p);
        if u==1
            x(3,iter) = T(3,u);
        end
    
    end


     % ---- Calculating the derivative values ----
    % A 14-vector, D, for 
    %           - D(1:4) = 1st derivatives (1,2,3,4)
    %           - D(5:8) = single-variable 2nd derivatives (1,2,3,4)
    %           - D(9:14) = cross-variable 2nd derivatives (12,13,14,23,24,34)

    % 1st derivatives
    D(1) = (T(3,2) - T(3,3))/(2*h(1));
    D(2) = (T(3,4) - T(3,5))/(2*h(2));

    % Single variable 2nd derivative
    D(5) = (T(3,2) + T(3,3) - T(3,1))/(h(1)^2);
    D(6) = (T(3,4) + T(3,5) - T(3,1))/(h(2)^2);

    % Cross variable 2nd derivative
    D(7) = (T(3,6) - T(3,7))/(2*h(1)*h(2)) - D(2)/h(1);

    % ---- Constructing the Gradient Vector ----
    G = zeros(2,1);
    G = [D(1),D(2)]';

    % ---- Constructing the Hessian matrix ----
    H = zeros(2,2);
    H(1,1) = D(5); H(2,2) = D(6);
    H(1,2) = D(7);
    H(2,1) = H(1,2);

    % ---- Moving to a new point in parameter space ----
    % Points are recorded in 5 x Imax matrix, X, where Imax = max number of iterations
    if iter < Imax
       d(1:2,iter) = H^(-1)*G;
       x(1:2,iter+1) = x(1:2,iter) - gamma*d(1:2,iter);
    end
    
    % Display the iteration number
    iter
end

losses_bestrn(1:2) = x(1:2,Imax)'; 